/*
 * 3-D gear wheels.  This program is in the public domain.
 *
 * Brian Paul
 */

/* Conversion to GLUT by Mark J. Kilgard */

/* Conversion to GtkGLExt by Naofumi Yasufuku */

/*  by Naofumi Yasufuku */

/* Frame Buffer Objects and Vertex Shader by Vladimir Bashkardin */

#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include <gtk/gtkgl.h>

#ifdef G_OS_WIN32
#define WIN32_LEAN_AND_MEAN 1
#include <windows.h>
#endif

#include <GL/gl.h>
#include <GL/glu.h>

#include "gtkglext_fbo_sl.h"

/*
 * Draw a gear wheel.  You'll probably want to call this function when
 * building a display list since we do a lot of trig here.
 *
 * Input:  inner_radius - radius of hole at center
 * outer_radius - radius at center of teeth
 * width - width of gear
 * teeth - number of teeth
 * tooth_depth - depth of tooth
 */

static void
gear(GLfloat inner_radius,
     GLfloat outer_radius,
     GLfloat width,
     GLint   teeth,
     GLfloat tooth_depth)
{
  GLint i;
  GLfloat r0, r1, r2;
  GLfloat angle, da;
  GLfloat u, v, len;

  r0 = inner_radius;
  r1 = outer_radius - tooth_depth / 2.0;
  r2 = outer_radius + tooth_depth / 2.0;

  da = 2.0 * G_PI / teeth / 4.0;

  glShadeModel(GL_FLAT);

  glNormal3f(0.0, 0.0, 1.0);

  /* draw front face */
  glBegin(GL_QUAD_STRIP);
  for (i = 0; i <= teeth; i++) {
    angle = i * 2.0 * G_PI / teeth;
    glVertex3f(r0 * cos(angle), r0 * sin(angle), width * 0.5);
    glVertex3f(r1 * cos(angle), r1 * sin(angle), width * 0.5);
    if (i < teeth) {
      glVertex3f(r0 * cos(angle), r0 * sin(angle), width * 0.5);
      glVertex3f(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da), width * 0.5);
    }
  }
  glEnd();

  /* draw front sides of teeth */
  glBegin(GL_QUADS);
  da = 2.0 * G_PI / teeth / 4.0;
  for (i = 0; i < teeth; i++) {
    angle = i * 2.0 * G_PI / teeth;

    glVertex3f(r1 * cos(angle), r1 * sin(angle), width * 0.5);
    glVertex3f(r2 * cos(angle + da), r2 * sin(angle + da), width * 0.5);
    glVertex3f(r2 * cos(angle + 2 * da), r2 * sin(angle + 2 * da), width * 0.5);
    glVertex3f(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da), width * 0.5);
  }
  glEnd();

  glNormal3f(0.0, 0.0, -1.0);

  /* draw back face */
  glBegin(GL_QUAD_STRIP);
  for (i = 0; i <= teeth; i++) {
    angle = i * 2.0 * G_PI / teeth;
    glVertex3f(r1 * cos(angle), r1 * sin(angle), -width * 0.5);
    glVertex3f(r0 * cos(angle), r0 * sin(angle), -width * 0.5);
    if (i < teeth) {
      glVertex3f(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da), -width * 0.5);
      glVertex3f(r0 * cos(angle), r0 * sin(angle), -width * 0.5);
    }
  }
  glEnd();

  /* draw back sides of teeth */
  glBegin(GL_QUADS);
  da = 2.0 * G_PI / teeth / 4.0;
  for (i = 0; i < teeth; i++) {
    angle = i * 2.0 * G_PI / teeth;

    glVertex3f(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da), -width * 0.5);
    glVertex3f(r2 * cos(angle + 2 * da), r2 * sin(angle + 2 * da), -width * 0.5);
    glVertex3f(r2 * cos(angle + da), r2 * sin(angle + da), -width * 0.5);
    glVertex3f(r1 * cos(angle), r1 * sin(angle), -width * 0.5);
  }
  glEnd();

  /* draw outward faces of teeth */
  glBegin(GL_QUAD_STRIP);
  for (i = 0; i < teeth; i++) {
    angle = i * 2.0 * G_PI / teeth;

    glVertex3f(r1 * cos(angle), r1 * sin(angle), width * 0.5);
    glVertex3f(r1 * cos(angle), r1 * sin(angle), -width * 0.5);
    u = r2 * cos(angle + da) - r1 * cos(angle);
    v = r2 * sin(angle + da) - r1 * sin(angle);
    len = sqrt(u * u + v * v);
    u /= len;
    v /= len;
    glNormal3f(v, -u, 0.0);
    glVertex3f(r2 * cos(angle + da), r2 * sin(angle + da), width * 0.5);
    glVertex3f(r2 * cos(angle + da), r2 * sin(angle + da), -width * 0.5);
    glNormal3f(cos(angle), sin(angle), 0.0);
    glVertex3f(r2 * cos(angle + 2 * da), r2 * sin(angle + 2 * da), width * 0.5);
    glVertex3f(r2 * cos(angle + 2 * da), r2 * sin(angle + 2 * da), -width * 0.5);
    u = r1 * cos(angle + 3 * da) - r2 * cos(angle + 2 * da);
    v = r1 * sin(angle + 3 * da) - r2 * sin(angle + 2 * da);
    glNormal3f(v, -u, 0.0);
    glVertex3f(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da), width * 0.5);
    glVertex3f(r1 * cos(angle + 3 * da), r1 * sin(angle + 3 * da), -width * 0.5);
    glNormal3f(cos(angle), sin(angle), 0.0);
  }

  glVertex3f(r1 * cos(0), r1 * sin(0), width * 0.5);
  glVertex3f(r1 * cos(0), r1 * sin(0), -width * 0.5);

  glEnd();

  glShadeModel(GL_SMOOTH);

  /* draw inside radius cylinder */
  glBegin(GL_QUAD_STRIP);
  for (i = 0; i <= teeth; i++) {
    angle = i * 2.0 * G_PI / teeth;
    glNormal3f(-cos(angle), -sin(angle), 0.0);
    glVertex3f(r0 * cos(angle), r0 * sin(angle), -width * 0.5);
    glVertex3f(r0 * cos(angle), r0 * sin(angle), width * 0.5);
  }
  glEnd();

}

static GdkGL_GL_EXT_framebuffer_object *fbo = NULL; /* Frame buffer object extension */ 
static GdkGL_GL_ARB_shader_objects *shaders = NULL; /* GL Shader Language extension */

static GLuint fb = 0;
static GLuint depth_rb = 0;
static GLuint fbo_tex = 0;
static GLuint fbo_tex_size = 0;

static GLint scraps_list = 0;

static GLhandleARB waving_program;
static GLhandleARB waving_vertex_shader;
static GLint waving_angle_uniform;

static GLfloat waving_angle = 0.f;
static GLfloat waving_speed = 1.0f;

static GLfloat width_gl = 0.f;
static GLfloat height_gl = 0.f;

static const GLchar *waving_vertex_source = " \n\
uniform float waving_angle; \n\
 \n\
void main (void) \n\
{ \n\
  vec4 v = vec4 (gl_Vertex.x, gl_Vertex.y, gl_Vertex.z, 1.0); \n\
 \n\
  v.z  = sin (3.14159 * gl_Vertex.x + radians (waving_angle)); \n\
  v.z += sin (3.14159 * gl_Vertex.y + radians (waving_angle)); \n\
  v.z *= 0.75; \n\
 \n\
  gl_Position = gl_ModelViewProjectionMatrix * v; \n\
 \n\
  gl_TexCoord[0] = gl_MultiTexCoord0; \n\
  gl_FrontColor = gl_Color; \n\
} \n\
";

static char *error_no_fbo = "Renderer does not support frame buffer objects";
static char *error_no_shaders = "Renderer does not support GL shader language";
static char *error_no_rect_tex = "Renderer does not support rectangle textures extension";

static GLfloat view_rotx = 20.0, view_roty = 30.0, view_rotz = 0.0;
static GLint gear1, gear2, gear3;
static GLfloat angle = 0.0;

static GTimer *timer = NULL;
static gint frames = 0;
static gfloat last_seconds = 0.f;

static gboolean is_sync = TRUE;

static gboolean
draw (GtkWidget      *widget,
      GdkEventExpose *event,
      gpointer        data)
{
  GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);

  /*** OpenGL BEGIN ***/
  if (!gdk_gl_drawable_gl_begin (gldrawable, glcontext))
    return FALSE;

  /*** Render to Frame Buffer Object ***/
  fbo->glBindFramebufferEXT (GL_FRAMEBUFFER_EXT, fb);

  glViewport (0, 0, fbo_tex_size, fbo_tex_size);
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  glFrustum (-1.0, 1.0, -1.0, 1.0, 5.0, 60.0);
  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();
  glTranslatef (0.0, 0.0, -40.0);

  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glPushMatrix ();
    glRotatef (view_rotx, 1.0, 0.0, 0.0);
    glRotatef (view_roty, 0.0, 1.0, 0.0);
    glRotatef (view_rotz, 0.0, 0.0, 1.0);

    glPushMatrix ();
      glTranslatef (-3.0, -2.0, 0.0);
      glRotatef (angle, 0.0, 0.0, 1.0);
      glCallList (gear1);
    glPopMatrix ();

    glPushMatrix ();
      glTranslatef (3.1, -2.0, 0.0);
      glRotatef (-2.0 * angle - 9.0, 0.0, 0.0, 1.0);
      glCallList (gear2);
    glPopMatrix ();

    glPushMatrix ();
      glTranslatef (-3.1, 4.2, 0.0);
      glRotatef (-2.0 * angle - 25.0, 0.0, 0.0, 1.0);
      glCallList (gear3);
    glPopMatrix ();

  glPopMatrix ();

  /*** Render to the screen ***/
  fbo->glBindFramebufferEXT (GL_FRAMEBUFFER_EXT, 0);

  glViewport (0, 0, widget->allocation.width, widget->allocation.height);
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  glFrustum (-width_gl / 2.0, width_gl / 2.0, -height_gl / 2.0, height_gl / 2.0, 5.0, 60.0);
  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();
  glTranslatef (0.0, 0.0, -7.0);

  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glPushAttrib (GL_ENABLE_BIT);

  glDisable (GL_LIGHTING);
  glDisable (GL_CULL_FACE);
  glDisable (GL_DEPTH_TEST);

  glEnable (GL_TEXTURE_RECTANGLE_ARB);

  glBindTexture (GL_TEXTURE_RECTANGLE_ARB, fbo_tex);

  shaders->glUseProgramObjectARB (waving_program);

  waving_angle += waving_speed;

  while (waving_angle > 360.0f) waving_angle -= 360.0f;
  while (waving_angle < 0.0f) waving_angle += 360.0f;

  shaders->glUniform1fARB (waving_angle_uniform, waving_angle);

  glCallList (scraps_list);

  shaders->glUseProgramObjectARB (0);

  glBindTexture (GL_TEXTURE_RECTANGLE_ARB, 0);

  glDisable (GL_TEXTURE_RECTANGLE_ARB);

  glPopAttrib ();

  if (gdk_gl_drawable_is_double_buffered (gldrawable))
    gdk_gl_drawable_swap_buffers (gldrawable);
  else
    glFlush ();

  gdk_gl_drawable_gl_end (gldrawable);
  /*** OpenGL END ***/

  frames++;

  {
    gdouble seconds = g_timer_elapsed (timer, NULL);
    if (seconds >= 5.0) {
      gdouble fps = frames / seconds;
      g_print ("%d frames in %6.3f seconds = %6.3f FPS\n", frames, seconds, fps);
      frames = 0;
      g_timer_reset (timer);
    }
  }

  return TRUE;
}

/* Tile the gears */
static void
tile_gears ()
{
  guint i, j;
  guint scrap_width_num = 50;
  guint scrap_height_num = scrap_width_num;
  GLfloat scrap_width = fbo_tex_size / (GLfloat)scrap_width_num;
  GLfloat scrap_height = fbo_tex_size / (GLfloat)scrap_height_num;
  GLfloat scrap_width_gl;
  GLfloat scrap_height_gl;
  GLfloat scrap_width_start;
  GLfloat scrap_height_start;

  if (width_gl < height_gl)
    {
       scrap_width_start = -width_gl / 2.0;
       scrap_height_start = -width_gl / 2.0;
       scrap_width_gl = width_gl / (GLfloat)scrap_width_num;
       scrap_height_gl = width_gl / (GLfloat)scrap_height_num;
    }
  else
    {
       scrap_width_start = -height_gl / 2.0;
       scrap_height_start = -height_gl / 2.0;
       scrap_width_gl = height_gl / (GLfloat)scrap_width_num;
       scrap_height_gl = height_gl / (GLfloat)scrap_height_num;
    }

  if (scraps_list)
    glDeleteLists (scraps_list, 1);

  scraps_list = glGenLists (1);
  glNewList (scraps_list, GL_COMPILE);
    /* Tile the rendered texture (in which the gears are depicted) */
    for (i = 0; i < scrap_width_num; i++) {
      for (j = 0; j < scrap_height_num; j++) {
        glBegin (GL_QUADS);
          glNormal3f (0, 0, -1);
          glTexCoord2d (scrap_width * i, scrap_height * j);
          glVertex2f (scrap_width_start + scrap_width_gl * i,
                      scrap_height_start + scrap_height_gl * j);
          glTexCoord2d (scrap_width * (i + 1), scrap_height * j);
          glVertex2f (scrap_width_start + scrap_width_gl * (i + 1),
                      scrap_height_start + scrap_height_gl * j);
          glTexCoord2d (scrap_width * (i + 1), scrap_height * (j + 1));
          glVertex2f (scrap_width_start + scrap_width_gl * (i + 1),
                      scrap_height_start + scrap_height_gl * (j + 1));
          glTexCoord2d (scrap_width * i, scrap_height * (j + 1));
          glVertex2f (scrap_width_start + scrap_width_gl * i,
                      scrap_height_start + scrap_height_gl * (j + 1));
        glEnd ();
      }
    }
  glEndList ();
}

/* get OpenGL extensions */
static void
init_extensions ()
{
  char *error_message = NULL;

  if (fbo)
    return;

  /*** Request OpenGL extensions ***/
  fbo = gdk_gl_get_GL_EXT_framebuffer_object ();
  if (fbo)
    {
      shaders = gdk_gl_get_GL_ARB_shader_objects ();
      if (shaders)
	    {
          if (FALSE == gdk_gl_query_gl_extension ("GL_ARB_texture_rectangle"))
            error_message = error_no_rect_tex;
        }
      else
        error_message = error_no_shaders;
    }
  else
    error_message = error_no_fbo;    

/*
 * Display message in a box and quit
 */
  if (error_message)
    {
	  GtkWidget *error_dialog = gtk_message_dialog_new (NULL,
                                                    	GTK_DIALOG_MODAL,
                                                    	GTK_MESSAGE_ERROR,
                                                    	GTK_BUTTONS_OK,
                                                    	"The graphics card sucks");
	  gtk_dialog_run (GTK_DIALOG (error_dialog));
	  gtk_widget_destroy (error_dialog);
	  exit (2);
	}
}

/* new GLSL program */
static void
build_GLSL_program ()
{

  if (NULL == shaders)
    init_extensions ();

  waving_program = shaders->glCreateProgramObjectARB ();
  waving_vertex_shader = shaders->glCreateShaderObjectARB (GL_VERTEX_SHADER_ARB);
  shaders->glShaderSourceARB (waving_vertex_shader, 1, &waving_vertex_source, NULL);
  shaders->glCompileShaderARB (waving_vertex_shader);
  shaders->glAttachObjectARB (waving_program, waving_vertex_shader);
  shaders->glLinkProgramARB (waving_program);

  GLcharARB infoLog[500];
  shaders->glGetInfoLogARB (waving_program, 500, NULL, infoLog);
  g_printf ("Waving vertex shader info log:\n");
  g_printf ("%s\n", infoLog);

  waving_angle_uniform = shaders->glGetUniformLocationARB (waving_program, "waving_angle");
}

/* new Frame Buffer Object */
static void
rebuild_fbo (GtkWidget *widget)
{

  if (NULL == fbo)
    init_extensions ();

  if (fb)
    {
      glDeleteTextures (1, &fbo_tex);
      fbo->glDeleteFramebuffersEXT (1, &fb);
      fbo->glDeleteRenderbuffersEXT (1, &depth_rb);
      fbo->glBindFramebufferEXT (GL_FRAMEBUFFER_EXT, 0);
    }

  if (fbo)
    {

      if (widget->allocation.width >= widget->allocation.height)
        {
          height_gl = 2.0 * (GLfloat)widget->allocation.height / (GLfloat)widget->allocation.width;
          width_gl = 2.0;
          fbo_tex_size = widget->allocation.height;
        }
      else
        {
          width_gl = 2.0 * (GLfloat)widget->allocation.width / (GLfloat)widget->allocation.height;
          height_gl = 2.0;
          fbo_tex_size = widget->allocation.width;
        }

      /*** Generate frame buffer object ***/
      glEnable (GL_TEXTURE_RECTANGLE_ARB);
      fbo->glGenFramebuffersEXT (1, &fb);
      glGenTextures (1, &fbo_tex);
      fbo->glGenRenderbuffersEXT (1, &depth_rb);
      fbo->glBindFramebufferEXT (GL_FRAMEBUFFER_EXT, fb);
      /*** Build a texture for the color buffer ***/
      glBindTexture (GL_TEXTURE_RECTANGLE_ARB, fbo_tex);
      glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
      glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
      glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
      glTexImage2D (GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGB8, fbo_tex_size, fbo_tex_size,
                    0, GL_RGB, GL_INT, NULL);
      fbo->glFramebufferTexture2DEXT (GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_RECTANGLE_ARB,
                                      fbo_tex, 0);
      /*** Setup the depth buffer storage ***/
      fbo->glBindRenderbufferEXT (GL_RENDERBUFFER_EXT, depth_rb);
      fbo->glRenderbufferStorageEXT (GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT24,
                                     fbo_tex_size, fbo_tex_size);
      fbo->glFramebufferRenderbufferEXT (GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT,
                                         GL_RENDERBUFFER_EXT, depth_rb);

      glBindTexture (GL_TEXTURE_RECTANGLE_ARB, 0);
      glDisable (GL_TEXTURE_RECTANGLE_ARB);
      fbo->glBindFramebufferEXT (GL_FRAMEBUFFER_EXT, 0);
    }
}

/* new window size or exposure */
static gboolean
reshape (GtkWidget         *widget,
         GdkEventConfigure *event,
         gpointer           data)
{
  GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);

  GLfloat h = (GLfloat) (widget->allocation.height) / (GLfloat) (widget->allocation.width);

  /*** OpenGL BEGIN ***/
  if (!gdk_gl_drawable_gl_begin (gldrawable, glcontext))
    return FALSE;

  rebuild_fbo (widget);

  tile_gears ();

  gdk_gl_drawable_gl_end (gldrawable);
  /*** OpenGL END ***/

  return TRUE;
}

static void
init(GtkWidget *widget,
     gpointer   data)
{
  GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);

  static GLfloat pos[4] = {5.0, 5.0, 10.0, 0.0};
  static GLfloat red[4] = {0.8, 0.1, 0.0, 1.0};
  static GLfloat green[4] = {0.0, 0.8, 0.2, 1.0};
  static GLfloat blue[4] = {0.2, 0.2, 1.0, 1.0};

  /*** OpenGL BEGIN ***/
  if (!gdk_gl_drawable_gl_begin (gldrawable, glcontext))
    return;

  build_GLSL_program ();

  rebuild_fbo (widget);

  glLightfv (GL_LIGHT0, GL_POSITION, pos);
  glEnable (GL_CULL_FACE);
  glEnable (GL_LIGHTING);
  glEnable (GL_LIGHT0);
  glEnable (GL_DEPTH_TEST);

  glPixelStorei (GL_PACK_ALIGNMENT, 1);
  glPixelStorei (GL_UNPACK_ALIGNMENT, 1);
  glDisable (GL_BLEND);

  /* make the gears */
  gear1 = glGenLists (1);
  glNewList (gear1, GL_COMPILE);
    glMaterialfv (GL_FRONT, GL_AMBIENT_AND_DIFFUSE, red);
    gear (1.0, 4.0, 1.0, 20, 0.7);
  glEndList ();

  gear2 = glGenLists (1);
  glNewList (gear2, GL_COMPILE);
    glMaterialfv (GL_FRONT, GL_AMBIENT_AND_DIFFUSE, green);
    gear (0.5, 2.0, 2.0, 10, 0.7);
  glEndList ();

  gear3 = glGenLists (1);
  glNewList (gear3, GL_COMPILE);
    glMaterialfv (GL_FRONT, GL_AMBIENT_AND_DIFFUSE, blue);
    gear (1.3, 2.0, 0.5, 10, 0.7);
  glEndList ();

  glEnable (GL_NORMALIZE);

  g_print ("\n");
  g_print ("GL_RENDERER   = %s\n", (char *) glGetString (GL_RENDERER));
  g_print ("GL_VERSION    = %s\n", (char *) glGetString (GL_VERSION));
  g_print ("GL_VENDOR     = %s\n", (char *) glGetString (GL_VENDOR));
  g_print ("GL_EXTENSIONS = %s\n", (char *) glGetString (GL_EXTENSIONS));
  g_print ("\n");

  gdk_gl_drawable_gl_end (gldrawable);
  /*** OpenGL END ***/

  /* create timer */
  if (timer == NULL)
    timer = g_timer_new ();

  g_timer_start (timer);
}

static gboolean
idle (GtkWidget *widget)
{
  angle += 0.2;

  /* Invalidate the whole window. */
  gdk_window_invalidate_rect (widget->window, &widget->allocation, FALSE);

  /* Update synchronously (fast). */
  if (is_sync)
    gdk_window_process_updates (widget->window, FALSE);

  return TRUE;
}

static guint idle_id = 0;

static void
idle_add (GtkWidget *widget)
{
  if (idle_id == 0)
    {
      idle_id = g_idle_add_full (GDK_PRIORITY_REDRAW,
                                 (GSourceFunc) idle,
                                 widget,
                                 NULL);
    }
}

static void
idle_remove (GtkWidget *widget)
{
  if (idle_id != 0)
    {
      g_source_remove (idle_id);
      idle_id = 0;
    }
}

static gboolean
map (GtkWidget   *widget,
     GdkEventAny *event,
     gpointer     data)
{
  idle_add (widget);

  return TRUE;
}

static gboolean
unmap (GtkWidget   *widget,
       GdkEventAny *event,
       gpointer     data)
{
  idle_remove (widget);

  return TRUE;
}

static gboolean
visible (GtkWidget          *widget,
         GdkEventVisibility *event,
         gpointer            data)
{
  if (event->state == GDK_VISIBILITY_FULLY_OBSCURED)
    idle_remove (widget);
  else
    idle_add (widget);

  return TRUE;
}

/* change view angle, exit upon ESC */
static gboolean
key (GtkWidget   *widget,
     GdkEventKey *event,
     gpointer     data)
{
  switch (event->keyval)
    {
    case GDK_z:
      view_rotz += 5.0;
      break;
    case GDK_Z:
      view_rotz -= 5.0;
      break;
    case GDK_Up:
      view_rotx += 5.0;
      break;
    case GDK_Down:
      view_rotx -= 5.0;
      break;
    case GDK_Left:
      view_roty += 5.0;
      break;
    case GDK_Right:
      view_roty -= 5.0;
      break;
    case GDK_Page_Up:
      if (waving_speed < 2.0f)
          waving_speed *= 2.0f;
      break;
    case GDK_Page_Down:
      if (waving_speed > 0.1f)
          waving_speed /= 2.0f;
      break;
    case GDK_Escape:
      gtk_main_quit ();
      break;
    default:
      return FALSE;
    }

  gdk_window_invalidate_rect (widget->window, &widget->allocation, FALSE);

  return TRUE;
}

int
main (int   argc,
      char *argv[])
{
  GdkGLConfig *glconfig;
  GtkWidget *window;
  GtkWidget *vbox;
  GtkWidget *drawing_area;
  GtkWidget *button;
  int i;

  /*
   * Init GTK.
   */

  gtk_init (&argc, &argv);

  /*
   * Init GtkGLExt.
   */

  gtk_gl_init (&argc, &argv);

  /*
   * Command line options.
   */

  for (i = 0; i < argc; i++)
    {
      if (strcmp (argv[i], "--async") == 0)
        is_sync = FALSE;
    }

  /*
   * Configure OpenGL-capable visual.
   */

  /* Try double-buffered visual */
  glconfig = gdk_gl_config_new_by_mode (GDK_GL_MODE_RGB    |
                                        GDK_GL_MODE_DEPTH  |
                                        GDK_GL_MODE_DOUBLE);
  if (glconfig == NULL)
    {
      g_print ("*** Cannot find the double-buffered visual.\n");
      g_print ("*** Trying single-buffered visual.\n");

      /* Try single-buffered visual */
      glconfig = gdk_gl_config_new_by_mode (GDK_GL_MODE_RGB   |
                                            GDK_GL_MODE_DEPTH);
      if (glconfig == NULL)
        {
          g_print ("*** No appropriate OpenGL-capable visual found.\n");
          exit (1);
        }
    }

  /*
   * Top-level window.
   */

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window), "gears");

  /* Get automatically redrawn if any of their children changed allocation. */
  gtk_container_set_reallocate_redraws (GTK_CONTAINER (window), TRUE);

  g_signal_connect (G_OBJECT (window), "delete_event",
                    G_CALLBACK (gtk_main_quit), NULL);

  /*
   * VBox.
   */

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (window), vbox);
  gtk_widget_show (vbox);

  /*
   * Drawing area for drawing OpenGL scene.
   */

  drawing_area = gtk_drawing_area_new ();
  gtk_widget_set_size_request (drawing_area, 300, 300);

  /* Set OpenGL-capability to the widget. */
  gtk_widget_set_gl_capability (drawing_area,
                                glconfig,
                                NULL,
                                TRUE,
                                GDK_GL_RGBA_TYPE);

  gtk_widget_add_events (drawing_area,
                         GDK_VISIBILITY_NOTIFY_MASK);

  g_signal_connect_after (G_OBJECT (drawing_area), "realize",
                          G_CALLBACK (init), NULL);
  g_signal_connect (G_OBJECT (drawing_area), "configure_event",
                    G_CALLBACK (reshape), NULL);
  g_signal_connect (G_OBJECT (drawing_area), "expose_event",
                    G_CALLBACK (draw), NULL);
  g_signal_connect (G_OBJECT (drawing_area), "map_event",
                    G_CALLBACK (map), NULL);
  g_signal_connect (G_OBJECT (drawing_area), "unmap_event",
                    G_CALLBACK (unmap), NULL);
  g_signal_connect (G_OBJECT (drawing_area), "visibility_notify_event",
                    G_CALLBACK (visible), NULL);

  g_signal_connect_swapped (G_OBJECT (window), "key_press_event",
                            G_CALLBACK (key), drawing_area);

  gtk_box_pack_start (GTK_BOX (vbox), drawing_area, TRUE, TRUE, 0);

  gtk_widget_show (drawing_area);

  /*
   * Simple quit button.
   */

  button = gtk_button_new_with_label ("Quit");

  g_signal_connect (G_OBJECT (button), "clicked",
                    G_CALLBACK (gtk_main_quit), NULL);

  gtk_box_pack_start (GTK_BOX (vbox), button, FALSE, FALSE, 0);

  gtk_widget_show (button);

  /*
   * Show window.
   */

  gtk_widget_show (window);

  /*
   * Main loop.
   */

  gtk_main ();

  return 0;
}
